# zbot/__init__.py
#
#

""" . """

__version__ = 50

## IMPORTS 

from zbot.utils import *

import collections
import threading
import logging
import _thread
import socket
import queue
import types
import errno
import select
import fcntl
import uuid
import json
import time
import imp
import sys
import os
import re

## ERRORS

class Error(BaseException): pass
class OverloadError(Error): pass
class MissingArgument(Error): pass
class MissingOutFunction(Error): pass
class NoText(Error): pass
class NoDate(Error): pass
class NoPath(Error): pass
class NoJSON(Error): pass
class NoFileName(Error): pass
class SignatureError(Error): pass
class NoSuchBotType(Error): pass
class NoEvent(Error): pass
class RemoteDisconnect(Error): pass
class NotOne(Error): pass
class TargetMissing(Error): pass

## BASE

class Object(dict):

    def __init__(zelf, *args, **kwargs): dict.__init__(zelf, *args, **kwargs)

    def __getattribute__(zelf, *args, **kwargs):
        name = args[0]
        result = ""
        if name == "date": return zelf.Date or zelf.added or ""
        if name == "time": return time_stamp(zelf.Date or zelf.date or zelf.added)
        if name == "what": result = get_cls(zelf)
        if name == "modname": result = get_modname(zelf)
        if name == "plugname": result = get_plugname(3)
        if name == "func": result = get_func(5)
        if name == "how": result =  get_how(2)
        if name == "kind": result = mj(zelf.modname, zelf.what, zelf.func)
        if name == "stdin": return sys.stdin
        if name == "stdout": return sys.stdout
        if result: zelf[name] = result ; return zelf[name]
        return dict.__getattribute__(zelf, *args, **kwargs)

    def __getattr__(zelf, name):
        try: return zelf[name]
        except KeyError:
            if name == "_result": zelf["_result"] = Object()
            if name == "_status": zelf["_status"] = Object()
            if name == "_ready": zelf["_ready"] = threading.Event()
        try: return zelf[name]
        except KeyError: return ""

    def __setattr__(zelf, name, value): return dict.__setitem__(zelf, name, value)

    def __exists__(zelf, a):
        try: return zelf[a]
        except KeyError: False

    def __lt__(zelf, a): return time_stamp(zelf.ddate) < time_stamp(a.ddate)

    ## TOUCH

    def touch(zelf, *args, **kwargs):
        try: res = getattr(zelf, args[0])
        except (TypeError, AttributeError): res = None
        return res

    ## locators
 
    def get_root(zelf, *args, **kwargs): return j(os.path.expanduser("~"), *args)

    def get_roots(zelf, *args, **kwargs):
        result = []
        return result

    def get_target(zelf):
        if "_target" in zelf: return zelf._target
        return None

    ## output

    def make_json(zelf, *args, **kwargs): return json.dumps(zelf.reduced(), default=smooth, *args, **kwargs)

    def make_full(zelf, *args, **kwargs): return json.dumps(zelf, default=smooth, *args, **kwargs)

    def make_signature(zelf, sig=None): return str(hashlib.sha1(bytes(str(sig or zelf), "utf-8")).hexdigest())

    def make_path(zelf, *args, **kwargs): path = zelf.get_path(*args, **kwargs) ; make_dir(path) ; return path

    ## input

    def load(zelf, *args, **kwargs):
        if args: path = zelf.make_path(*args, **kwargs)
        else: path = zelf.make_path()
        return zelf.load_file(path)

    def load_file(zelf, *args, **kwargs):
        path = args[0]
        ondisk = zelf.read(path) 
        fromdisk = json.loads(ondisk)
        if "data" in fromdisk: zelf.update(fromdisk["data"])
        return zelf

    def load_json(zelf, *args, **kwargs): zelf.update(json.loads(args[0]))

    def read(zelf, *args, **kwargs):
        logging.info("read %s" % args[0])
        path = args[0]
        try: f = open(path, "r")
        except IOError as ex:
            if ex.errno == errno.ENOENT: return "{}"
            raise
        if zelf.do_test: f.line_buffering = False
        res = ""
        for line in f:
            if not line.strip().startswith("#"): res += line
        if not res.strip(): return "{}"
        f.close()
        return res

    ## persistence

    def dump(zelf, *args, **kwargs):
        # object attributes settins before save
        path = args[0]
        todisk = Object()
        todisk.data = zelf.reduced()
        todisk.data.date = zelf.date
        todisk.how = zelf.how
        todisk.create_type = zelf.what
        todisk.stime = time.ctime(time.time())
        todisk.modname = zelf.modname
        todisk.version = __version__
        todisk.signature = zelf.make_signature()
        try: result = todisk.make_json(indent=2, ensure_ascii=False)
        except TypeError: raise NoJSON()
        return result

    def save(zelf, *args, **kwargs):
        if not args: t = rtime()
        else: t = args[0]
        for path in kernel.cfg.rootlist:
            root = zelf.get_root(path, t)
            make_dir(root)
            zelf.sync(root)
 
    def sync(zelf, *args, **kwargs):
        # syn the object to disk
        path = args[0]
        fn = path.split(os.sep)[-1]
        todisk = zelf.dump(path, **kwargs)
        datafile = open(path + ".tmp", 'w')
        fcntl.flock(datafile, fcntl.LOCK_EX | fcntl.LOCK_NB)
        datafile.write(headertxt % (fn, __version__, "%s characters" % len(todisk)))
        datafile.write(todisk)
        datafile.write("\n")
        fcntl.flock(datafile, fcntl.LOCK_UN)
        datafile.close()
        os.rename(path + ".tmp", path)
        logging.warn("saved %s" % path)
        return zelf

    ## state

    def register(zelf, *args, **kwargs):
        name = args[0]
        obj = args[1]
        zelf[name] = obj

    def reduced(zelf, *args, **kwargs):
        res = Object()
        for key in zelf.keys():
            k = str(key)
            if k.startswith("_"): continue   
            if key in ["args", "rest", "first"]: continue
            res[key] = zelf[key]
        return res

    def pretty(zelf): return json.dumps(zelf, indent=2)

    ## objectors

    def objects(zelf, *args, **kwargs):
        path = zelf.get_root(kernel.cfg.workdir, *args, **kwargs)
        flist = list_files(path, **kwargs)
        res = []
        for fnn in flist:
            try: obj = Object().load_file(fnn)
            except ValueError: logging.warn("no json in %s" % fnn) ; continue
            res.append(obj)
        logging.warn("%s objects found" % len(res))
        return res

    ## result

    def parse(zelf, *args, **kwargs):
        try: zelf.ucmnd, zelf.rest = zelf.txt.split(" ", 1)
        except: zelf.ucmnd = zelf.txt ; zelf.rest = ""
        if zelf.ucmnd and zelf.ucmnd[0] == ".": zelf.ucmnd = zelf.ucmnd[1:]
        return zelf

    def prepare(zelf, *args, **kwargs):
        zelf.opts = get_opts(zelf.txt)
        zelf.args = get_args(zelf.txt)
        zelf.knobs = get_knobs(zelf.txt)
        zelf.parse()
        return zelf

    def ready(zelf): zelf._ready.set()

    def wait(zelf, sec=180.0): zelf._ready.wait(sec) 

    def add(zelf, value): zelf._result[time.time()] = value 

    def reply(zelf, txt): zelf.say(zelf.channel, txt)

    def show(zelf): return ["%s=%s" % (a, b) for a, b in zelf.items() if b]

    def _raw(zelf, txt): zelf._target._raw(txt)

    def say(zelf, *args, **kwargs):
        if zelf.outer: o = zelf.outer ; o.write(str(args[1])) ; o.write("\n") ; o.flush()
        else: zelf._target.say(*args, **kwargs)

    def display(zelf, *args, **kwargs):
        for res in sorted(zelf._result.keys()): d = zelf._result[res] ; zelf.say(zelf.channel or "", d)
             
    def put(zelf, *args, **kwargs): zelf._target.put(*args, **kwargs)

    def handle(zelf, *args , **kwargs): res = resolve(zelf, *args, **kwargs) ; return res

## TASKS

class Runner(threading.Thread):

    def __init__(zelf, *args, **kwargs):
        threading.Thread.__init__(zelf, None, zelf._loop, "thread.%s" % str(time.time()), args, kwargs)
        zelf.setDaemon(True)
        zelf._queue = queue.Queue()
        zelf._outqueue = queue.Queue()
        zelf._state = "idle"

    def _loop(zelf, *args, **kwargs):
        logging.warn("start %s" % str(zelf))
        while zelf._state in ["running", "idle", "callback", "once"]:
            o = zelf._queue.get()
            if zelf._state == "stop": break
            try:
                func = o.a[0]
                o.returned = func(*o.a[1:], **o.kw)
            except: error()
            time.sleep(0.001)
        logging.warn("stop %s" % str(zelf))

    def stop(zelf, *args, **kwargs): zelf._state = "stop"

    def put(zelf, *args, **kwargs):
        o = Object()
        o.a = args
        o.kw = kwargs
        zelf._queue.put(o)

class TaskRunner(Runner):

    def _loop(zelf, *args, **kwargs):
        logging.warn("start %s" % str(zelf))
        while zelf._state in ["running", "idle", "callback", "once"]:
            o = zelf._queue.get()
            if zelf._state == "stop": break
            logging.info("run %s" % str(o))
            kernel.cb.handle_cb(o.a[0])
            time.sleep(0.001)
        logging.warn("stop %s" % str(zelf))

class Spider(Runner):

    def __init__(zelf, *args, **kwargs):
        Runner.__init__(zelf, *args, **kwargs)
        zelf.errors = []
        zelf.urls = []
        zelf.url = Object()
        zelf.followed = []
        zelf.speed = 0
        zelf.depth = 5

    def crawl(zelf, *args, **kwargs):
        zelf.start()
        zelf.get_urls(*args, **kwargs)

    def get_urls(zelf, *args, **kwargs):
        url = args[0]
        urls = []
        if not zelf.url:
           zelf.url.url = url
           zelf.url.basepath, zelf.url.base, zelf.url.root, zelf.url.file = parse_url(zelf.url.url)
        pnr = len(url.split("/"))
        if pnr >  zelf.depth: logging.warn("%s depth > 5" % url) ; return
        if url not in zelf.urls:
            zelf.urls.append(url)
            content = do_url("GET", url)
            newurl = need_redirect(content)
            if newurl: content = do_url("GET", newurl) ; logging.warn("redirecting to %s" % newurl)
            newurl2 = need_redirect(content)
            if newurl2: content = do_url("GET", url) ; logging.warn("redirecting to %s" % newurl2)
            time.sleep(zelf.speed)
            zelf.speed += 0.1
            urls = parse_urls(url, content.read()) 
            o = Object()
            o.spider = True
            o.orig_url = url
            o.urls = urls
            o.save()
            for u in urls:
                if u in zelf.urls: continue
                if not zelf.url.base in u: continue
                if u in zelf.errors: continue
                zelf.put(zelf.get_urls, u)
        return urls

## DISPATCH

class Dispatcher(Object):

    max = 50
    runners = collections.deque() 

    def stop(zelf, name=None):
        for runner in zelf.runners:
            if name and name in runner.name: runner.stop()

    def put(zelf, *args, **kwargs):
        if not args: raise NoTask()
        target = zelf.get_target()
        target.put(*args, **kwargs)
        return args[0]

    def get_target(zelf):
        target = None
        for runner in zelf.runners:
            if runner._queue and runner._state == "idle": target = runner
        if not target: target = zelf.makenew()
        return target

    def makenew(zelf, *args, **kwargs):
        if len(zelf._runners) < zelf.max:
            runner = TaskRunner(*args, **kwargs)
            runner.start()
            zelf.runners.append(runner)
        else: runner = random.choice(zelf._runners)
        return runner

    def cleanup(zelf, dojoin=False):
        todo = []
        for runner in zelf.runners:
            if runner.stopped or not len(runner.queue): todo.append(runner)
        for runner in todo: runner.stop() ; zelf.runners.remove(runner)

## CALLBACKS

class Callbacks(Dispatcher):

    def register(zelf, cbtype, cb, *args, **kwargs):
        logging.debug("register %s" % cbtype)
        if cbtype not in zelf: zelf[cbtype] = []
        zelf[cbtype].append(cb)

    def handle_cb(zelf, *args, **kwargs):
        event = args[0]
        logging.info(event)
        target = get_cls(event)
        logging.info("cb %s %s" % (target, event.show()))
        functions = []
        try: functions.extend(zelf[target])
        except KeyError: pass
        for func in functions:
            try: pre = getattr(func, "pre")
            except AttributeError: pre = None
            if pre and not pre(event): logging.debug("pre %s" % str(func)) ; return
            try: result = func(event)
            except: error()
        return event

## BOTS

class Bot(Callbacks): 

    def connect(zelf, *args, **kwargs): pass

    def exit(zelf, *args, **kwargs):
        try: sys.stdout.flush() 
        except Exception as ex: logging.warn(str(ex))
        sys.stdout.close()
        os._exit(0)

    def get_one(zelf): return None

    def read_some(zelf, *args, **kwargs): pass

    def put_some(zelf, *args, **kwargs):
        input = args[0]
        [i, o, e] = select.select([input,], [], [], 0.05)
        gotcha = False
        if i: text = i[0].read()
        return gotcha

    def get_prompt(zelf, *args, **kwargs): return ""

    def begin(zelf, *args, **kwargs):
        zelf.connect()
        if not zelf._state: zelf._state = "running"
        zelf.run_forever(*args, **kwargs)
        
    def run_forever(zelf, *args, **kwargs):
        while zelf._state in ["running", "idle", "once", "callback"]:
            time.sleep(0.001)
            e = zelf.get_one()
            if e:
                zelf.do_one(e)
                if e._state == "once": break
        zelf.ready()
 
    def do_one(zelf, *args, **kwargs):
        ok = True
        try: event = args[0] ; event._target = zelf ; zelf.put(event) ; event.wait() 
        except KeyboardInterrupt: raise
        except IndexError: ok = False
        except Exception: error() ; ok = False
        return ok

## PLUGINS

class Plugins(Object):

    def get_names(zelf, plugsdir): return [x[:-3] for x in os.listdir(plugsdir) if x.endswith(".py")]

    def load_plugs(zelf, path):
        logging.warn("plugs %s" % path)
        for plugname in zelf.get_names(path):
            if "__" in plugname: continue
            try: mod = zelf.load_mod(plugname, path, force=True)
            except: error() ; continue

    def load_package(zelf, modname):
        mod = __import__(modname)
        path, fn = os.path.split(mod.__file__)
        path += os.sep + "plugs"
        zelf.load_plugs(path)        

    def load_mod(zelf, plugname, pdir="", force=False):
        logging.info("load %s - %s" % (zelf.what.lower(), plugname))
        if plugname in zelf:
            if not force: return zelf[plugname]
            zelf[plugname] = imp.reload(zelf[plugname])
        else:
            if not pdir: pdir = j(zelf.root, "plugs")
            search = imp.find_module(plugname, [pdir,])
            zelf[plugname] = imp.load_module(plugname, *search)
        zelf.plug_exec(plugname, "init")
        return zelf[plugname]

    def plug_exec(zelf, plugname, item):
        try: todo = getattr(zelf[plugname], item) ; todo() ; logging.info("exec %s" % get_name(todo))
        except AttributeError: pass

    def unload(zelf, plugname):
        zelf.plug_exec(plugname, "shutdown")
        del zelf[plugname]

    def reload(zelf, plugname, force=False):
        zelf.unload(plugname)
        mod = zelf.load_mod(plugname, force)
        return mod

## BOOT

def boot(*args, **kwargs):
    global kernel
    cfg = args[0]
    kernel.cfg.update(cfg)
    if cfg.do_shell: hello("ZBOT")
    set_workdir(kernel.cfg.workdir)
    if not kernel.cfg.loglevel: cfg.loglevel = "error"
    from zbot.log import log_config
    log_config(cfg.loglevel)
    if kernel.cfg.loglevel:
        logging.warn("E G G S")
        logging.warn("")
        show_eggs()
    if kernel.cfg.loglevel:
        logging.warn("")
        logging.warn("C O N F I G")
        logging.warn("")
        for line in kernel.cfg.show(): logging.warn(line)
    if kernel.cfg.loglevel:
        logging.warn("")
        logging.warn("B O O T")
        logging.warn("")
    kernel.cb.register("Object", handle_loop)
    return kernel

## HANDLER

def handle_loop(event):
    try: event.prepare() ; event.handle() ; event.ready()
    except KeyboardInterrupt: raise

## WORKDIR

def set_workdir(*args, **kwargs):
    global kernel
    workdir = args[0]
    if not workdir: workdir = "zbot.workdir"
    wd = workdir.split(os.sep)[-1]
    if not wd: wd = workdir.split(os.sep)[-2] 
    else: wd = "zbot.workdir"
    if not kernel.cfg.workdir: kernel.cfg.workdir = j(os.getcwd(), wd)
    kernel.cfg.rootlist = []
    kernel.cfg.rootlist.append(kernel.cfg.workdir)
    kernel.cfg.rootlist.append(j(os.getcwd(), ".zelf", wd, day(), ""))
    kernel.cfg.rootlist.append(j("Public", wd, day(), ""))
    kernel.cfg.rootlist.append(j("Dropbox", "backups", wd, day(), ""))
    make_dir(kernel.cfg.workdir)

## VARS

kernel = Object()
kernel.cfg = Object()
kernel.plugs = Plugins()
kernel.cb = Callbacks()
kernel.cmnds = Callbacks()
kernel.tests = Callbacks()
