# zbot/drivers/xmpp.py
#
#

""" XMPP bot build on sleekxmpp. """

## IMPORTS

from zbot.utils import error, stripped, attributes, subelements
from zbot import Bot, Object, handle_loop

import collections
import threading
import _thread
import logging
import getpass
import types
import time
import sys
import os
import re

## XMPPBot

class XMPPBot(Bot):

    def __init__(zelf, *args, **kwargs):
        import sleekxmpp
        from sleekxmpp import clientxmpp
        Bot.__init__(zelf, *args, **kwargs)
        zelf.port = zelf.port or 5222
        zelf.xmpp = clientxmpp.ClientXMPP(zelf.user, getpass.getpass())
        zelf.xmpp.add_event_handler("session_start", zelf.session_start)
        zelf.xmpp.add_event_handler("message", zelf.handle_message)
        zelf.xmpp.add_event_handler('disconnected', zelf.handle_disconnected)
        zelf.xmpp.add_event_handler('connected', zelf.handle_connected)
        zelf.xmpp.add_event_handler('presence_available', zelf.handle_presence)
        zelf.xmpp.add_event_handler('presence_dnd', zelf.handle_presence)
        zelf.xmpp.add_event_handler('presence_xa', zelf.handle_presence)
        zelf.xmpp.add_event_handler('presence_chat', zelf.handle_presence)
        zelf.xmpp.add_event_handler('presence_away', zelf.handle_presence)   
        zelf.xmpp.add_event_handler('presence_unavailable', zelf.handle_presence) 
        zelf.xmpp.add_event_handler('presence_subscribe', zelf.handle_presence)   
        zelf.xmpp.add_event_handler('presence_subscribed', zelf.handle_presence)  
        zelf.xmpp.add_event_handler('presence_unsubscribe', zelf.handle_presence) 
        zelf.xmpp.add_event_handler('presence_unsubscribed', zelf.handle_presence)
        zelf.xmpp.add_event_handler('groupchat_message', zelf.handle_message)
        zelf.xmpp.add_event_handler('groupchat_presence', zelf.handle_presence)
        zelf.xmpp.add_event_handler('groupchat_subject', zelf.handle_presence)
        zelf.xmpp.add_event_handler('failed_auth', zelf.handle_failedauth)
        zelf.xmpp.exception = zelf.exception
        zelf.xmpp.use_signals()
        if zelf.openfire:
            import ssl
            zelf.xmpp.ssl_version = ssl.PROTOCOL_SSLv3
        zelf._connected = threading.Event()

    def _raw(zelf, txt): zelf.xmpp.send_raw(txt)

    def connect(zelf): zelf.xmpp.connect((zelf.server, zelf.port), use_ssl=zelf.openfire)

    def session_start(zelf, event): zelf.xmpp.send_presence()
        
    def exception(zelf, ex): logging.error(str(ex))

    def handle_failedauth(zelf, error, *args): logging.warn(error) 

    def handle_failure(zelf, ex, *args, **kwargs): logging.info(str(ex))

    def handle_disconnected(zelf, *args, **kwargs): zelf._connected.clear() ; logging.error("disconnected")

    def handle_connected(zelf, *args, **kwargs): zelf._connected.set() ; logging.warn("connected!")

    def loop(zelf, *args, **kwargs): zelf.xmpp.process(block=True)

    def say(zelf, *args, **kwargs): zelf.xmpp.send_message(args[0], args[1])

    def get_one(zelf, *args, **kwargs): return

    def handle_message(zelf, data, *args, **kwargs):
        if '<delay xmlns="urn:xmpp:delay"' in str(data): logging.debug("ignoring delayed message") ; return
        logging.warn("< %s" % data)
        m = Object(**data)
        if m.type == "error": logging.warn("error %s" % m.make_json()) ; return
        m.origin = stripped(str(m["from"]))
        m.channel = m.origin
        m.to = m.origin
        m.element = "message"
        try: m.txt = m["body"]
        except KeyError: pass
        m._target = zelf
        zelf.put(m)

    def handle_presence(zelf, data, *args, **kwargs):
        logging.info("< %s" % data)
        p = Object(**data)
        if "from" in p and p["from"]:
            p.origin = stripped(str(p["from"]))
            p.to = p.origin
        else: p.origin = "" ; p.to = ""
        p.element = "presence"
        if p.type == 'subscribe':
            pres = Object({'to': p["from"], 'type': 'subscribed'})
            zelf.send(pres)
            pres = Object({'to': p["from"], 'type': 'subscribe'})
            zelf.send(pres)
        p.no_dispatch = True
        p._target = zelf
        zelf.put(p)
